/*****************************************************
Project : networkswitch
Version : 3.0
Author  : Kalim Moghul
Company : Cornell University

Chip type           : ATmega16
Clock frequency     : 16.000000 MHz
*****************************************************/

#include <mega16.h>
#include <delay.h>
#include <stdio.h>

#include "ringbuffer.h"
#define UBYTE unsigned char
#define UINT  unsigned int

#define BTx   0b10000000
#define BAck  0b01000000
#define PTx   0b00100000
#define PAck  0b00010000
#define Dmask 0b11110011

#define RBUFSIZE 24

#pragma regalloc-
UINT rbuf[16][RBUFSIZE];          /* allocate ringbuffer storage per tag/address pair */
ringbuffer rb[16];                 /* allocate ringbuffer per dest tag/address pair */
#pragma regalloc+

void proc0(void);
void proc1(void);
void proc2(void);
void proc3(void);

void main(void)
{
    int i;
    
    //PORTn Pinouts
    //Pin SignalDirection
    //7   BTx   OUT
    //6   BAck  OUT
    //5   PReq  IN
    //4   PAck  IN
    //3   Data3 OUT
    //2   Data2 OUT
    //1   Data1 IN
    //0   Data0 IN
    DDRA=0xcc;//interface for node 0
    PORTA=0x00;
    DDRB=0xcc;  //interface for node 1
    PORTB=0x00;
    DDRC=0xcc;  //interface for node 2
    PORTC=0x00;
    DDRD=0xcc;  //interface for node 3
    PORTD=0x00;

    //USART Init for debugging 
    //UCSRB = 0x18;
    //UBRRL = 103;
  
    delay_ms(1000);
  
    //Initialize ring buffers
    for (i = 0; i < 16; i++)
        rbinit(rb + i, rbuf[i], RBUFSIZE);                 //init ring buffers
  
    while(1)
    {
        proc0();
        proc1();
        proc2(); 
        proc3();
    }//end main loop
}


#define COMPORT PORTA
#define COMPIN  PINA
void proc0()
{
    UBYTE procnum = 0;    //this function's owner (should be const, but CodeVisionAVR doesn't allow it)
    static UINT Raddrdata, Taddrdata;  //concatenated address and data, as stored in ringbuffer
    static UBYTE Raddr, Rtag, Rdata, Tdata, Taddr;
    static UBYTE rxstate = 0, txstate = 6, txq = 0, txready = 0, RAck = 0, TAck = PAck, rxpause = 0;
    UBYTE i;
    
    //**********
    // receive
    //**********
    if (((COMPIN & PTx) != RAck) || rxpause)
    {
        RAck = (rxpause) ? RAck : RAck ^ PTx;
        if (rxstate == 0)
        {
            Raddr = COMPIN & 0x3;                                //read dest address crumb
            COMPORT = COMPORT ^ BAck;                              //toggle the Ack
            rxstate++;
        }
        else if (rxstate == 1)
        {
            Rtag = COMPIN & 0x3;                                  //read dest tag crumb
            COMPORT = COMPORT ^ BAck;
            rxstate++;
        }
        else if (rxstate <= 5)
        {
            Rdata = (Rdata << 2) | (COMPIN & 0x3);                //read data crumb
            COMPORT = COMPORT ^ BAck;                               //toggle the Ack
            if (rxstate == 5)   //last data received, so construct value for ringbuffer
            {
                Raddrdata = (((UINT)procnum) << 8) | (UINT)Rdata;
            }
            rxstate++;
        }
        if (rxstate == 6)
        {
            rxpause = !(rbput(rb + ((Rtag<<2)|Raddr), Raddrdata));  //try to put addr/data in queue, stall if full
            if (!rxpause)
            {
                rxstate = 0;
            }
        }
    }
    
    //**********
    // transmit
    //**********
    if (txstate == 6)   //if ready to transmit something
    {
        txready = 0;
        for (i = 0; i < 4; i++) //round robin arbitration to select from which dest tag queue to transmit
        {
            txq = (txq + 1) & 0x3;                      //select next dest tag for this processor
            if (!rbisempty(rb + ((txq<<2)|procnum)))    //see if it has anything in queue with selected tag
            {
                txready = 1;
                txstate = 0;
                Taddrdata = *(rbget(rb + ((txq<<2)|procnum))); //get addr/data value
                Taddr = (UBYTE)(Taddrdata >> 8);
                Tdata = (UBYTE)(Taddrdata);
                break;
            }
        }
    }
    
    if (txready)
    {
        if ((COMPIN & PAck) != TAck)
        {
            TAck = TAck ^ PAck;
            if (txstate == 0)
            {
                COMPORT = (COMPORT & Dmask) | (Taddr << 2);                   //put source addr crumb
                txstate++;
            }
            else if (txstate == 1)
            {
                COMPORT = (COMPORT & Dmask) | (txq << 2);                     //put dest tag crumb
                txstate++;
            }
            else if (txstate <= 4)
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata >> (8 - (txstate << 1))) & 0xc);   //put data crumb
                txstate++;
            }
            else
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata << 2) & 0xc);  //put data crumb
                txstate = 6;
            }
            COMPORT = COMPORT ^ BTx;                                //toggle the transmit signal
        }
    }
}
#undef COMPORT
#undef COMPIN

#define COMPORT PORTB
#define COMPIN  PINB
void proc1()
{
    UBYTE procnum = 1;    //this function's owner (should be const, but CodeVisionAVR doesn't allow it)
    static UINT Raddrdata, Taddrdata;  //concatenated address and data, as stored in ringbuffer
    static UBYTE Raddr, Rtag, Rdata, Tdata, Taddr;
    static UBYTE rxstate = 0, txstate = 6, txq = 0, txready = 0, RAck = 0, TAck = PAck, rxpause = 0;
    UBYTE i;
    
    //**********
    // receive
    //**********
    if (((COMPIN & PTx) != RAck) || rxpause)
    {
        RAck = (rxpause) ? RAck : RAck ^ PTx;
        if (rxstate == 0)
        {
            Raddr = COMPIN & 0x3;                                //read dest address crumb
            COMPORT = COMPORT ^ BAck;                              //toggle the Ack
            rxstate++;
        }
        else if (rxstate == 1)
        {
            Rtag = COMPIN & 0x3;                                  //read dest tag crumb
            COMPORT = COMPORT ^ BAck;
            rxstate++;
        }
        else if (rxstate <= 5)
        {
            Rdata = (Rdata << 2) | (COMPIN & 0x3);                //read data crumb
            COMPORT = COMPORT ^ BAck;                               //toggle the Ack
            if (rxstate == 5)   //last data received, so construct value for ringbuffer
            {
                Raddrdata = (((UINT)procnum) << 8) | (UINT)Rdata;
            }
            rxstate++;
        }
        if (rxstate == 6)
        {
            rxpause = !(rbput(rb + ((Rtag<<2)|Raddr), Raddrdata));  //try to put addr/data in queue, stall if full
            if (!rxpause)
            {
                rxstate = 0;
            }
        }
    }
    
    //**********
    // transmit
    //**********
    if (txstate == 6)   //if ready to transmit something
    {
        txready = 0;
        for (i = 0; i < 4; i++) //round robin arbitration to select from which dest tag queue to transmit
        {
            txq = (txq + 1) & 0x3;                      //select next dest tag for this processor
            if (!rbisempty(rb + ((txq<<2)|procnum)))    //see if it has anything in queue with selected tag
            {
                txready = 1;
                txstate = 0;
                Taddrdata = *(rbget(rb + ((txq<<2)|procnum))); //get addr/data value
                Taddr = (UBYTE)(Taddrdata >> 8);
                Tdata = (UBYTE)(Taddrdata);
                break;
            }
        }
    }
    
    if (txready)
    {
        if ((COMPIN & PAck) != TAck)
        {
            TAck = TAck ^ PAck;
            if (txstate == 0)
            {
                COMPORT = (COMPORT & Dmask) | (Taddr << 2);                   //put source addr crumb
                txstate++;
            }
            else if (txstate == 1)
            {
                COMPORT = (COMPORT & Dmask) | (txq << 2);                     //put dest tag crumb
                txstate++;
            }
            else if (txstate <= 4)
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata >> (8 - (txstate << 1))) & 0xc);   //put data crumb
                txstate++;
            }
            else
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata << 2) & 0xc);  //put data crumb
                txstate = 6;
            }
            COMPORT = COMPORT ^ BTx;                                //toggle the transmit signal
        }
    }
}
#undef COMPORT
#undef COMPIN

#define COMPORT PORTC
#define COMPIN  PINC
void proc2()
{
    UBYTE procnum = 2;    //this function's owner (should be const, but CodeVisionAVR doesn't allow it)
    static UINT Raddrdata, Taddrdata;  //concatenated address and data, as stored in ringbuffer
    static UBYTE Raddr, Rtag, Rdata, Tdata, Taddr;
    static UBYTE rxstate = 0, txstate = 6, txq = 0, txready = 0, RAck = 0, TAck = PAck, rxpause = 0;
    UBYTE i;
    
    //**********
    // receive
    //**********
    if (((COMPIN & PTx) != RAck) || rxpause)
    {
        RAck = (rxpause) ? RAck : RAck ^ PTx;
        if (rxstate == 0)
        {
            Raddr = COMPIN & 0x3;                                //read dest address crumb
            COMPORT = COMPORT ^ BAck;                              //toggle the Ack
            rxstate++;
        }
        else if (rxstate == 1)
        {
            Rtag = COMPIN & 0x3;                                  //read dest tag crumb
            COMPORT = COMPORT ^ BAck;
            rxstate++;
        }
        else if (rxstate <= 5)
        {
            Rdata = (Rdata << 2) | (COMPIN & 0x3);                //read data crumb
            COMPORT = COMPORT ^ BAck;                               //toggle the Ack
            if (rxstate == 5)   //last data received, so construct value for ringbuffer
            {
                Raddrdata = (((UINT)procnum) << 8) | (UINT)Rdata;
            }
            rxstate++;
        }
        if (rxstate == 6)
        {
            rxpause = !(rbput(rb + ((Rtag<<2)|Raddr), Raddrdata));  //try to put addr/data in queue, stall if full
            if (!rxpause)
            {
                rxstate = 0;
            }
        }
    }
    
    //**********
    // transmit
    //**********
    if (txstate == 6)   //if ready to transmit something
    {
        txready = 0;
        for (i = 0; i < 4; i++) //round robin arbitration to select from which dest tag queue to transmit
        {
            txq = (txq + 1) & 0x3;                      //select next dest tag for this processor
            if (!rbisempty(rb + ((txq<<2)|procnum)))    //see if it has anything in queue with selected tag
            {
                txready = 1;
                txstate = 0;
                Taddrdata = *(rbget(rb + ((txq<<2)|procnum))); //get addr/data value
                Taddr = (UBYTE)(Taddrdata >> 8);
                Tdata = (UBYTE)(Taddrdata);
                break;
            }
        }
    }
    
    if (txready)
    {
        if ((COMPIN & PAck) != TAck)
        {
            TAck = TAck ^ PAck;
            if (txstate == 0)
            {
                COMPORT = (COMPORT & Dmask) | (Taddr << 2);                   //put source addr crumb
                txstate++;
            }
            else if (txstate == 1)
            {
                COMPORT = (COMPORT & Dmask) | (txq << 2);                     //put dest tag crumb
                txstate++;
            }
            else if (txstate <= 4)
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata >> (8 - (txstate << 1))) & 0xc);   //put data crumb
                txstate++;
            }
            else
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata << 2) & 0xc);  //put data crumb
                txstate = 6;
            }
            COMPORT = COMPORT ^ BTx;                                //toggle the transmit signal
        }
    }
}
#undef COMPORT
#undef COMPIN

#define COMPORT PORTD
#define COMPIN  PIND
void proc3()
{
    UBYTE procnum = 3;    //this function's owner (should be const, but CodeVisionAVR doesn't allow it)
    static UINT Raddrdata, Taddrdata;  //concatenated address and data, as stored in ringbuffer
    static UBYTE Raddr, Rtag, Rdata, Tdata, Taddr;
    static UBYTE rxstate = 0, txstate = 6, txq = 0, txready = 0, RAck = 0, TAck = PAck, rxpause = 0;
    UBYTE i;
    
    //**********
    // receive
    //**********
    if (((COMPIN & PTx) != RAck) || rxpause)
    {
        RAck = (rxpause) ? RAck : RAck ^ PTx;
        if (rxstate == 0)
        {
            Raddr = COMPIN & 0x3;                                //read dest address crumb
            COMPORT = COMPORT ^ BAck;                              //toggle the Ack
            rxstate++;
        }
        else if (rxstate == 1)
        {
            Rtag = COMPIN & 0x3;                                  //read dest tag crumb
            COMPORT = COMPORT ^ BAck;
            rxstate++;
        }
        else if (rxstate <= 5)
        {
            Rdata = (Rdata << 2) | (COMPIN & 0x3);                //read data crumb
            COMPORT = COMPORT ^ BAck;                               //toggle the Ack
            if (rxstate == 5)   //last data received, so construct value for ringbuffer
            {
                Raddrdata = (((UINT)procnum) << 8) | (UINT)Rdata;
            }
            rxstate++;
        }
        if (rxstate == 6)
        {
            rxpause = !(rbput(rb + ((Rtag<<2)|Raddr), Raddrdata));  //try to put addr/data in queue, stall if full
            if (!rxpause)
            {
                rxstate = 0;
            }
        }
    }
    
    //**********
    // transmit
    //**********
    if (txstate == 6)   //if ready to transmit something
    {
        txready = 0;
        for (i = 0; i < 4; i++) //round robin arbitration to select from which dest tag queue to transmit
        {
            txq = (txq + 1) & 0x3;                      //select next dest tag for this processor
            if (!rbisempty(rb + ((txq<<2)|procnum)))    //see if it has anything in queue with selected tag
            {
                txready = 1;
                txstate = 0;
                Taddrdata = *(rbget(rb + ((txq<<2)|procnum))); //get addr/data value
                Taddr = (UBYTE)(Taddrdata >> 8);
                Tdata = (UBYTE)(Taddrdata);
                break;
            }
        }
    }
    
    if (txready)
    {
        if ((COMPIN & PAck) != TAck)
        {
            TAck = TAck ^ PAck;
            if (txstate == 0)
            {
                COMPORT = (COMPORT & Dmask) | (Taddr << 2);                   //put source addr crumb
                txstate++;
            }
            else if (txstate == 1)
            {
                COMPORT = (COMPORT & Dmask) | (txq << 2);                     //put dest tag crumb
                txstate++;
            }
            else if (txstate <= 4)
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata >> (8 - (txstate << 1))) & 0xc);   //put data crumb
                txstate++;
            }
            else
            {
                COMPORT = (COMPORT & Dmask) | ((Tdata << 2) & 0xc);  //put data crumb
                txstate = 6;
            }
            COMPORT = COMPORT ^ BTx;                                //toggle the transmit signal
        }
    }
}
#undef COMPORT
#undef COMPIN

